# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.

import os
os.environ["JUPYTER_PLATFORM_DIRS"] = "1"

#
# Release Gym Environments
#

from .baby_robot_interface import BabyRobotInterface
from .baby_robot_v0 import BabyRobot_v0



#
# Example Gym Environments
#

from .baby_robot_env_v0 import BabyRobotEnv_v0
from .baby_robot_env_v1 import BabyRobotEnv_v1
from .baby_robot_env_v2 import BabyRobotEnv_v2
from .baby_robot_env_v3 import BabyRobotEnv_v3
from .baby_robot_env_v4 import BabyRobotEnv_v4
from .baby_robot_env_v5 import BabyRobotEnv_v5
from .baby_robot_env_v6 import BabyRobotEnv_v6
from .baby_robot_env_v7 import BabyRobotEnv_v7
